/*
 * TGraphicSearch.java
 *
 * Created on 2005/06/27
 * Copyright 2005 Takeshi Yoneki / OSTRA / OSTRACISM CO., All rights reserved.
 */

package jp.maccreators.ostra.ggg4j;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.nio.channels.FileChannel;

import javax.swing.JFrame;

/**
 * TGraphicSearch 摜NX
 * @author	ostra
 */
public class TGraphicSearch {

	public interface Callback {
		public void hitItemCallback(TFileInfo info);
		public void progressCallback(TFileInfo info);
		public void searchEndCallback();
	}

	public Callback m_context = null;
	public JFrame m_parentWindow = null;
	public TFileInfo m_target = new TFileInfo("");
	public String m_topDirPath = "";
	public String m_indexFilePath = "";
	public int m_bits = 3;
	public int m_matchRate = 70;
	public int m_pixelNumber = 8;
	
	public static final int SEARCH_DIRECTORY = 0;
	public static final int SEARCH_INDEXFILE = 1;
	public int m_searchMethod = SEARCH_DIRECTORY;
	
	Thread m_work = null;
	boolean m_continue = false;

	/**
	 * ݒt@CT|[g
	 */
	public void loadSetting(TSetting setting) {
		m_target.m_filePath = setting.targetFilePath;
		m_topDirPath = setting.topDirPath;
		m_indexFilePath = setting.indexFilePath;
		m_bits = setting.bits;
		m_matchRate = setting.matchRate;
		m_pixelNumber = setting.pixelNumber;
		m_searchMethod = setting.searchMethod;
	}

	public void saveSetting(TSetting setting) {
		setting.targetFilePath = m_target.m_filePath;
		setting.topDirPath = m_topDirPath;
		setting.indexFilePath = m_indexFilePath;
		setting.bits = m_bits;
		setting.matchRate = m_matchRate;
		//setting.pixelNumber = m_pixelNumber;
		setting.searchMethod = m_searchMethod;
	}

	/**
	 * ƒ
	 */
	public boolean isWorking() {
		return m_work != null;
	}

	/**
	 * ~v
	 */
	public void stop() {
		synchronized(this) {
			m_continue = false;
		}
	}

	/**
	 * 
	 */
	public void join() {
		if (isWorking() == false)
			return;
		stop();
		try {
			m_work.join();
		}
		catch (Exception e) {
		}
	}

	/**
	 * 
	 */
	public boolean beginSearch() {
		boolean good = true;
		if (m_work != null)
			good = false;
		if (good)
			if (m_target == null)
				good = false;
		if (good)
			if (m_topDirPath == null && m_indexFilePath == null)
				good = false;
		if (good)
			if (m_searchMethod == SEARCH_INDEXFILE)
				good = checkIndexFile(m_indexFilePath);
		if (good)
			if (m_target.makeThumbnail(m_pixelNumber) == false)
				good = false;

		if (good == false) {
			if (m_context != null)
				m_context.searchEndCallback();
			return false;
		}

		m_continue = true;
		if (m_searchMethod == SEARCH_DIRECTORY)
			searchDirectoryThread();
		else if (m_searchMethod == SEARCH_INDEXFILE)
			searchIndexFileThread();
		return true;
	}

	/**
	 * fBNgXbh
	 */
	void searchDirectoryThread() {
		m_work = new Thread() {
			public void run() {
				m_target.makeThumbnail(m_pixelNumber);
				searchDirectorySub(m_topDirPath);
				if (m_context != null)
					m_context.searchEndCallback();
				m_work = null;
			}
	
			boolean searchDirectorySub(String dirPath) {
				File dir = new File(dirPath); 
				File[] files = dir.listFiles();
				int i;
				for (i = 0; i < files.length; i++) {
					yield();
					if (m_continue == false)
						return false;
					if (files[i].isFile()) {
						TFileInfo current = new TFileInfo(files[i].getPath());
						if (m_context != null)
							m_context.progressCallback(current);
						if (current.makeThumbnail(m_pixelNumber) == false)
							continue;
						if (m_target.isEqualThumbnail(m_bits, m_matchRate, current)) {
							if (m_context != null)
								m_context.hitItemCallback(current);
						}
					}
					else if (files[i].isDirectory()) {
						if (searchDirectorySub(files[i].getPath()) == false)
							return false;
					}
				}
				return true;			
			}
		};
		m_work.start();
	}

	/**
	 * CfbNXt@CXbh
	 */
	void searchIndexFileThread() {
		m_work = new Thread() {
			public void run() {
				try {
					File f = new File(m_indexFilePath);
					FileInputStream stream = new FileInputStream(f);
					if (stream != null) {
						TChunk ck = new TChunk();
						ck.read(stream);
						TChunkHead ckhd = new TChunkHead(ck);
						m_pixelNumber = ckhd.m_pixelNumber;
						m_target.makeThumbnail(m_pixelNumber);
						FileChannel channel = stream.getChannel();
						String filePath = "";
						for (;;) {
							if (channel.size() - channel.position() < 8) {
								break;
							}
							if (m_continue == false)
								break;
							ck = new TChunk();
							ck.read(stream);
							if (ck.m_ckID.equals(TChunk.CKID_PATH)) {
								TChunkPath ckpt = new TChunkPath(ck);
								filePath = ckpt.m_path;
							}
							else if (ck.m_ckID.equals(TChunk.CKID_THUMB)) {
								TChunkThumb cktb = new TChunkThumb(ck);
								if (filePath != null && filePath.length() > 0) {
									TFileInfo current = new TFileInfo();
									current.m_filePath = filePath;
									current.m_thumbnail = cktb.m_thumbnail;
									current.m_thumbnailCount = m_pixelNumber * m_pixelNumber;
									if (m_target.isEqualThumbnail(m_bits, m_matchRate, current)) {
										if (m_context!= null)
											m_context.hitItemCallback(current);
									}
								}
							}
							else {
							}
						}
						stream.close();
					}
				}
				catch (Exception e) {
				}
				if (m_context != null)
					m_context.searchEndCallback();
				m_work = null;
			}
		};
		m_work.start();
	}

	/**
	 * CfbNXt@C`FbN
	 */
	public boolean checkIndexFile(String path) {
		boolean result = false;
		try {
			File f = new File(path);
			FileInputStream stream = new FileInputStream(f);
			if (stream != null) {
				FileChannel channel = stream.getChannel();
				byte[] buff = new byte[4];
				if (stream.read(buff) == 4) {
					String ckID = new String(buff, 0, 4, "UTF-8");
					if (ckID.equals("GGHD")) {
						channel.position(0);
						TChunk ck = new TChunk();
						ck.read(stream);
						TChunkHead ckhd = new TChunkHead(ck);
						m_pixelNumber = ckhd.m_pixelNumber;
						result = true;
					}
				}
				stream.close();
			}
		}
		catch (Exception e) {
		}
		return result;
	}

	/**
	 * CfbNXt@C쐬
	 */
	public boolean makeIndexFile() {
		boolean good = true;
		if (m_work != null)
			good = false;
		if (good)
			if (m_topDirPath == null || m_topDirPath.length() == 0)
				good = false;

		if (good == false) {
			if (m_context != null)
				m_context.searchEndCallback();
			return false;
		}

		m_continue = true;
		makeIndexFileThread();
		return true;
	}

	/**
	 * CfbNXt@C쐬Xbh
	 */
	void makeIndexFileThread() {
		m_work = new Thread() {
			public void run() {
				FileOutputStream stream = null;
				try {
					stream = new FileOutputStream(new File(m_indexFilePath));
					if (stream != null) {
						TChunkHead ckhd = new TChunkHead();
						ckhd.m_pixelNumber = m_pixelNumber;
						ckhd.write(stream);
						makeIndexFileSub(m_topDirPath, stream);
						stream.close();
					}
				}
				catch(Exception e) {
				}
				if (m_context != null)
					m_context.searchEndCallback();
				m_work = null;
			}
			
			boolean makeIndexFileSub(String dirPath, OutputStream stream) {
				File dir = new File(dirPath); 
				File[] files = dir.listFiles();
				int i;
				for (i = 0; i < files.length; i++) {
					yield();
					if (m_continue == false)
						return false;
					if (files[i].isFile()) {
						TFileInfo current = new TFileInfo(files[i].getPath());
						if (m_context != null)
							m_context.progressCallback(current);
						if (current.makeThumbnail(m_pixelNumber) == false)
							continue;
						// 
						TChunkPath ckpt = new TChunkPath();
						ckpt.m_path = current.m_filePath;
						ckpt.write(stream);
						TChunkThumb cktb = new TChunkThumb();
						cktb.m_thumbnail = current.m_thumbnail;
						cktb.write(stream);
					}
					else if (files[i].isDirectory()) {
						if (makeIndexFileSub(files[i].getPath(), stream) == false)
							return false;
					}
				}
				return true;
			}
		};
		m_work.start();
	}

}
